package anywheresoftware.b4a.objects;

import java.io.File;
import java.util.Date;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import com.healthmarketscience.jackcess.Column;
import com.healthmarketscience.jackcess.Cursor;
import com.healthmarketscience.jackcess.DataType;
import com.healthmarketscience.jackcess.Database;
import com.healthmarketscience.jackcess.Index;
import com.healthmarketscience.jackcess.IndexCursor;
import com.healthmarketscience.jackcess.Table;
import com.healthmarketscience.jackcess.IndexData.ColumnDescriptor;

import anywheresoftware.b4a.BA.Author;
import anywheresoftware.b4a.BA.DependsOn;
import anywheresoftware.b4a.BA.Hide;
import anywheresoftware.b4a.BA.ShortName;
import anywheresoftware.b4a.BA.Version;

/**
 *The JackcessDatabase object represents an Access database. 
 *Once opened the tables in the database may be passed to the Initialize method of a JackcessTable object.
 */
@Version(1.0f)
@Author("Andrew Graham")
@ShortName("JackcessDatabase")
@DependsOn(values = { "jackcess-3.5.1", "commons-lang-2.6" })
public class JackcessDatabase
{
	// http://jackcess.sourceforge.net/
	// 1.0 	initial release
	private static final double version = 1.0;

	/**
	 *Returns the version of the library.
	 */
	public double getVersion()
	{
		return version;
	}

	private Database dbase;

	/**
	 *This library is based on Jackcess, a pure Java library for reading from and writing to MS Access databases.
	 *Presently Access database versions 2000-2010 are supported and Access 97 read-only.
	 *It will obviously help to have some knowledge of MS Access in order to understand it.
	 *Jackcess does not support SQL queries so all database manipulations are done procedurally.
	 *
	 *Documentation for Jackcess is rather limited but the Javadocs for Jackcess can be found here.
	 *<link>http://jackcess.sourceforge.net/apidocs/index.html|http://jackcess.sourceforge.net/apidocs/index.html</link>
	 *
	 *This library depends on commons-lang-2.6.jar and jackcess-1.2.10.jar. It also requires several files to be present
	 *in a myproject\res\raw folder. Those files may be copied from the demo in the archive and must be made read-only.
	 *
	 *Jackcess and therefore also this library is licensed under the GNU Lesser General Public License.
	 *<link>http://www.gnu.org/licenses/lgpl.html|http://www.gnu.org/licenses/lgpl.html</link>
	 *Copies of both the General Public License and Lesser General Public License are in the provided archive.
	 *
	 *Copyright 2013 Anywhere Software
	 */
	public void LIBRARY_DOC()
	{
	}

	public JackcessDatabase()
	{
		System.setProperty("com.healthmarketscience.jackcess.brokenNio=true", "");
		System.setProperty("com.healthmarketscience.jackcess.resourcePath=/res/raw/", "");
		Thread.currentThread().setContextClassLoader(Database.class.getClassLoader());
	}	
	
	/**
	 *Closes the presently opened Access database.
	 */
	public void Close() throws Exception
	{
		if (dbase == null)
			return;
		dbase.close();
		dbase = null;
	}	
	
	/**
	 *Helper method to turn the Java Date type returned from a database date field into a ticks value.
	 */
	public long DateToTicks(Date date)
	{
		return date.getTime();
	}
	
	/**
	 *Helper method to turn a ticks value into the Java Date type required for a database date field.
	 */
	public Date DateFromTicks(long ticks)
	{
		return new Date(ticks);
	}
	
	/**
	 *Returns a multiline String containing potentially useful details of the Database.
	 */
	public String Details()
	{
		return dbase.toString();
	}
	
	/**
	 *Returns the file format of this database as one of MSISAM, V1997, V2000, V2003, V2007 or V2010.  
	 */
	public String GetFileFormat() throws Exception
	{
		return dbase.getFileFormat().toString();
	}	
	
	/**
	 *Returns a reference to any available table in this access database, including system tables.
	 *Warning, this method is not designed for common use, only for the occassional time when access to
	 *a system table is necessary.
	 */
	public Table GetSystemTable(String systemtablename) throws Exception
	{
		return dbase.getSystemTable(systemtablename);
	}
	
	/**
	 *Returns a String array containing the names of all the system tables in the database.
	 */
	public String[] GetSystemTableNames() throws Exception
	{
		String[] sa = new String[0];
		sa = dbase.getSystemTableNames().toArray(sa);
		return sa;
	}
	
	/**
	 *Returns a Table for the database for use in Table.Initialize.
	 */
	public Table GetTable(String table) throws Exception
	{
		return dbase.getTable(table);
	}
	
	/**
	 *returns a String array containing the names of all the user tables in the database.
	 */
	public String[] GetTableNames() throws Exception
	{
		String[] sa = new String[0];
		sa = dbase.getTableNames().toArray(sa);
		return sa;
	}
	
	/**
	 *Flushes any current changes to the database file (and any linked databases) to disk.
	 */
	public void Flush() throws Exception
	{
		dbase.flush();
	}
	
	/**
	 *Opens the specified Access database.
	 */	public void Open(String filename) throws Exception
	{
		dbase = Database.open(new File(filename));		
	}	

	/**
	 *Returns True if a database is open.
	 */
	public boolean IsOpen()
	{
		return dbase != null;
	}

	
	//###########################################################################
	// JackcessTable
	//###########################################################################
	
	/**
	 *The JackcessTable object represent a table within the open database.
	 */
	@ShortName("JackcessTable")
	public static class JackcessTable
	{
		
		private Table table;
		private Map<String, Object> rowmap;
		
		/**
		 *Initialises this object with a database table provided by JackcessDatabase,GetTable.
		 */
		public void Initialize(Table table)
		{
			this.table = table;			
		}		
		
		/**
		 *Adds a single row to this table and writes it to disk.
		 *The values are expected to be given in the order that the Columns are listed by getColumns() method.
		 *values are the row values for a single row. The given row array will be modified if this table contains
		 *an auto-number column, otherwise it will not be modified. 
		 *Any Date values should be passed as ticks to the Jackcess.DateFromTicks helper method
		 *and the returned Date object used here.
		 */
		public void AddRow(Object[] values) throws Exception
		{
			table.addRow(values);
		}
		
		/**
		 *Add multiple rows to this table, only writing to disk after all rows have been written,
		 *and every time a data page is filled. This is much more efficient than calling addRow multiple times. 
		 *Note, if this table has an auto-number column, the values written will be put back into the given row arrays.
		 *Any Date values should be passed as ticks to the Jackcess.DateFromTicks helper method
		 *and the returned Date object used here.
		 */
		public void AddRows(List<Object[]> values) throws Exception
		{
			table.addRows(values);
		}
		
		/**
		 *Delete the current row (retrieved by a call to GetNextRow).  
		 * @throws Exception 
		 */
		public void DeleteCurrentRow() throws Exception
		{
			table.deleteCurrentRow();
		}
		
		/**
		 *Returns a string containing potentially useful details of the Table.
		 */
		public String Details()
		{
			return table.toString();
		}
		
		/**
		 *Returns a string representation of the entire table in tab-delimited format.
		 */
		public String Display() throws Exception
		{
			String s = table.display();
			table.reset();
			return s;
		}
		
		/**
		 *Returns the names of each of the columns in the table as a String array.
		 */
		public String[] GetColumnNames()
		{
			List<Column> cols = table.getColumns();
			String[] cnames = new String[cols.size()];
			int i = 0;
			for (Column col : cols)
			{
				cnames[i] = col.getName();
				i++;
			}
			return cnames;
		}
		
		/**
		 *Returns the Access data type of each of the columns in the table as a String array.
		 */
		public String[] GetColumnDataTypes()
		{
			 List<Column> cols = table.getColumns();
			 String[] cnames = new String[cols.size()];
			 int i = 0;
			 for (Column col : cols)
			 {
				 cnames[i] = col.getType().toString();
				 i++;
			 }
			 return cnames;
		}
		
		/**
		 *Returns the Java data type of each of the columns in the table as a String array.
		 *Note that the Date type needs special treatment when read from the database.
		 *There are two helper methods, DateToTicks and DatefromTicks in JackcessDatabase to amp
		 *database Date type values to and from ticks for use in Basic4android.
		 *Access types map to Java types as follows.
		 *
		 *	BINARY: Byte[] of max length 255 bytes
		 *	BOOLEAN: Boolean
		 *	BYTE: Byte
		 *	COMPLEX_TYPE: Integer - an autonumber field which is the key for a secondary table which holds the data
		 *	DOUBLE: Double
		 *	FLOAT: Float
		 *	GUID: String with the pattern {xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx}
		 *	INT: Short
		 *	LONG: Integer
		 *	MEMO: String of max length 8388607 chars
		 *	MONEY: BigDecimal with at most 4 decimal places.
		 *	NUMERIC: BigDecimal 
		 *	OLE: Byte[] of max length 16777215 bytes.
		 *	SHORT_DATE_TIME: Date - basic4android does not directly support this type
		 *	TEXT: String of max length 255 chars				 
		 */
		public String[] GetColumnJavaTypes()
		{
			 List<Column> cols = table.getColumns();
			 String[] ctypes = new String[cols.size()];
			 int i = 0;
			 for (Column col : cols)
			 {
				 DataType dt = col.getType();
				 switch(dt)
				 {
					 case BINARY: ctypes[i] = "Byte[]"; break;
					 case BOOLEAN: ctypes[i] = "Boolean"; break;
					 case BYTE: ctypes[i] =  "Byte"; break;
					 case COMPLEX_TYPE: ctypes[i] = "Integer"; break;
					 case DOUBLE: ctypes[i] = "Double"; break;
					 case FLOAT: ctypes[i] = "Float"; break;
					 case GUID: ctypes[i] = "String"; break;
					 case INT: ctypes[i] = "Short"; break;
					 case LONG: ctypes[i] = "Integer"; break;
					 case MEMO: ctypes[i] = "String"; break;
					 case MONEY: ctypes[i] = "BigDecimal"; break;
					 case NUMERIC: ctypes[i] = "BigDecimal"; break;
					 case OLE: ctypes[i] = "Byte[]"; break;
					 case SHORT_DATE_TIME: ctypes[i] = "Date"; break;
					 case TEXT: ctypes[i] = "String"; break;				 
					 default:  ctypes[i] = "Unmapped type: " + dt.toString(); break;				 
				 }
				 i++;
			 }
			 return ctypes;
		}		
		
		/**
		 *Most Access data types map to types directly supported by Basic4android and can be returned as is.
		 *Only dates need special treatment to turn them into ticks for the Basic4android DateTime object.
		 *Rather than put a check for a Date in every call to this method it is incumbent on the programmer
		 *to pass the Date object to the JackcessDatabase.DateToTicks helper method.
		 */
		public Object GetColumnValue(String columnname)
		{
			return rowmap.get(columnname);
		}
		
		/**
		 *Returns an Index for the table for use in JackcessIndex.Initialize.
		 */public Index GetIndex(String indexname)
		{
			return table.getIndex(indexname);			
		}
		 
		/**
		 *Returns the internal wrapped Table object.
		 */
		public Table GetTable()
		{
			return table;
		}
		
		/**
		 *Returns the names of each of the indexes in the table as a String array.
		 */
		public String[] GetIndexNames()
		{
			List<Index> indexes = table.getIndexes();
			String[] inames = new String[indexes.size()];
			int i = 0;
			for (Index idx : indexes)
			{
				inames[i] = idx.getName();
				i++;
			}
			return inames;			
		}
		
		/**
		 *Reads the next row into an internal Map containing a set of Column name (String) and value (Object) pairs.
		 *The values of individual columns in the row may be obtained by GetColumnValue or GetColumnValueAndDate
		 */
		public void GetNextRow() throws Exception
		{
			rowmap = (LinkedHashMap<String, Object>) table.getNextRow();
		}		
		
		/**
		 *Returns the number of rows in the table.
		 */
		public int GetRowCount()
		{
			return table.getRowCount();
		}		
		
		/**
		 *After calling this method, GetNextRow will return the first row in the table, 
		 */
		public void Reset()
		{
			table.reset();
		}

		/**
		 *Update the current row with the provided values.
		 *Any Date values should be passed as ticks to the Jackcess.DateFromTicks helper method
		 *and the returned Date object used here.
		 *
		 *Note the after invoking this method GetColumnValue will return the old values not the new ones.
		 *This is unlike the behaviour of Cursor.UpdateCurrentRow and Cursor.SetCurrentRowValue after which
		 *the new values ARE available. The reason for this is that Cursor rereads the current row to update
		 *its stored values but a Table cannot do this as it lacks the necessary method.
		 */
		public void UpdateCurrentRow(Object[] values) throws Exception
		{
			table.updateCurrentRow(values);
		}
		
	}
	
	//###########################################################################
	// JackcessIndex
	//###########################################################################
	
	
	@ShortName("JackcessIndex")
	public static class JackcessIndex
	{		
		private Index index;
		
		/**
		 *Initialises this object with a database index provided by JackcessTable.GetIndex.
		 */
		public void Initialize(Index index)
		{
			this.index = index;			
		}		
		
		/**
		 *Returns a string containing potentially useful details of the Index.
		 */
		public String Details()
		{
			return index.toString();
		}
		
		/**
		 *Returns the names of each of the columns in the index as a String array.
		 */
		public String[] GetColumnNames()
		{
			List<ColumnDescriptor> cols = index.getColumns();
			String[] cnames = new String[cols.size()];
			int i = 0;
			for (ColumnDescriptor col : cols)
			{
				cnames[i] = col.getName();
				i++;
			}
			return cnames;
		}		
		
		/**
		 *Returns the internal wrapped Index object.
		 */
		public Index GetIndex()
		{
			return index;
		}
		
		/**
		 *Returns true if the index is a foreign key.
		 *i.e. it is a field in the table that is a primary key in another table
		 */
		public boolean IsForeignKey()
		{
			return index.isForeignKey();
		}
		
		/**
		 *Returns True if the index is a primary key.
		 *i.e.  it is a field whose values uniquely identify each record in that table
		 */
		public boolean IsPrimaryKey()
		{
			return index.isPrimaryKey();
		}
		
		/**
		 *Returns True if the index entries must be unique values.
		 *Some notes about uniqueness: 
		 *Access does not seem to consider multiple null entries invalid for a unique index.
		 *Text indexes collapse case, and Access seems to compare only the index entry bytes,
		 *therefore two strings which differ only in case will violate the unique constraint.
		 */
		public boolean IsUnique()
		{
			return index.isUnique();
		}		
		
	}
	
	//###########################################################################
	// JackcessIndexCursor
	//###########################################################################
	
	/**
	 *An index cursor provides efficient navigation of a table using a predefined index column or columns.
	 */
	@ShortName("JackcessIndexCursor")
	public static class JackcessIndexCursor extends JackcessCursor
	{		
		
		/**
		 *Initialises this object with a table provided by JackcessDatabase.GetTable or JackcessTable.GetTable
		 *and a database index provided by JackcessTable.GetIndex or JackcessIndex.GetIndex.
		 */
		public void Initialize(Table table, Index index) throws Exception
		{
			this.cursor = IndexCursor.createCursor(table, index);			
		}
		
		/**
		 *Moves to the first row (as defined by the cursor) where the index entries are >= the given values.
		 *If an exception is thrown, the cursor is restored to its previous state. 
		 *
		 *entryValues is an array with the column values for the index's columns.
		 */		
		public void FindClosestRowByEntry(Object[] entryvalues) throws Exception
		{
			((IndexCursor)cursor).findClosestRowByEntry(entryvalues);
			rowmap = (LinkedHashMap<String, Object>) cursor.getCurrentRow();
		}
		
		/**
		 *Moves to the first row (as defined by the cursor) where the index entries match the given values.
		 *If a match is not found (or an exception is thrown), the cursor is restored to its previous state.
		 *Warning, this method always starts searching from the beginning of the Table (you cannot use it to find successive matches).
		 *If the entry is found the row is read into an internal Map containing a set of Column name (String) and value (Object) pairs.
		 *The values of individual columns in the row may be obtained by GetColumnValue or GetColumnValueAndDate
		 *entryValues is an array with the column values for the index's columns.
		 *
		 */
		public boolean FindFirstRowByEntry(Object[] entryvalues) throws Exception
		{
			boolean found = ((IndexCursor)cursor).findFirstRowByEntry(entryvalues);
			if (found)
			{
				rowmap = (LinkedHashMap<String, Object>) cursor.getCurrentRow();
			}
			return found;
		}
	
	}
	
	//###########################################################################
	// JackcessCursor
	//###########################################################################
	
	/**
	 *A cursor provides navigation of a table.
	 */
	@ShortName("JackcessCursor")
	public static class JackcessCursor
	{
		@Hide
		protected Cursor cursor;
		@Hide
		protected Map<String, Object> rowmap;
		
		/**
		 *Initialises this object with a table provided by JackcessDatabase.GetTable or JackcessTable.GetTable.
		 */
		public void Initialize(Table table) throws Exception
		{
			this.cursor = Cursor.createCursor(table);			
		}
		
		/**
		 *Delete the current row. Throws IllegalStateException if the current row is not valid
		 *(at beginning or end of table), or already deleted 
		 */
		public void DeleteCurrentRow() throws Exception
		{
			cursor.deleteCurrentRow();
		}
				
		/**
		 *Moves to the first row (as defined by the cursor) where the given column has the given value.
		 *If a match is not found (or an exception is thrown), the cursor is restored to its previous state.
		 *Warning, this method always starts searching from the beginning of the Table (you cannot use it to find successive matches). 
		 */
		public boolean FindFirstRow(String column, Object value) throws Exception
		{
			boolean found = cursor.findFirstRow(cursor.getTable().getColumn(column), value);
			if (found)
			{
				rowmap = (LinkedHashMap<String, Object>) cursor.getCurrentRow();
			}
			return found;
		}
		
		/**
		 *Moves to the next row (as defined by the cursor) where the given column has the given value.
		 *If a match is not found (or an exception is thrown), the cursor is restored to its previous state. 
		 */
		public boolean FindNextRow(String column, Object value) throws Exception
		{
			boolean found = cursor.findNextRow(cursor.getTable().getColumn(column), value);
			if (found)
			{
				rowmap = (LinkedHashMap<String, Object>) cursor.getCurrentRow();
			}
			return found;
		}				
	
		/**
		 *Most Access data types map to types directly supported by Basic4android and can be returned as is.
		 *Only dates need special treatment to turn them into ticks for the Basic4android DateTime object.
		 *Rather than put a check for a Date in every call to this method it is incumbent on the programmer
		 *to pass the Date object to the JackcessDatabase.DateToTicks helper method.
		 */
		public Object GetColumnValue(String columnname)
		{
			return rowmap.get(columnname);
		}
		
		/**
		 *Returns the internal wrapped Cursor object.
		 */
		public Cursor GetCursor()
		{
			return cursor;
		}
		
		/**
		 *Reads the next row into an internal Map containing a set of Column name (String) and value (Object) pairs.
		 *The values of individual columns in the row may be obtained by GetColumnValue or GetColumnValueAndDate
		 */
		public void GetNextRow() throws Exception
		{
			rowmap = (LinkedHashMap<String, Object>) cursor.getNextRow();
		}
		
		/**
		 *Reads the previous row into an internal Map containing a set of Column name (String) and value (Object) pairs.
		 *The values of individual columns in the row may be obtained by GetColumnValue or GetColumnValueAndDate
		 */
		public void GetPreviousRow() throws Exception
		{
			rowmap = (LinkedHashMap<String, Object>) cursor.getPreviousRow();
		}
		
		/**
		 *Returns true if the cursor is currently positioned after the last row, false otherwise
		 */
		public boolean IsAfterLast() throws Exception
		{
			return cursor.isAfterLast();
		}
		
		/**
		 *Returns true if the cursor is currently positioned before the first row, false otherwise. 
		 */
		public boolean IsBeforeFirst() throws Exception
		{
			return cursor.isBeforeFirst();
		}
		
		/**
		 *Updates a single value in the current row.
		 *Any Date values should be passed as ticks to the Jackcess.DateFromTicks helper method
		 *and the returned Date object used here.
		 */
		public void SetCurrentRowValue(String column, Object value) throws Exception
		{
			cursor.setCurrentRowValue(cursor.getTable().getColumn(column), value);
			rowmap = cursor.getCurrentRow();
		}
		
		/**
		 *Update the current row with the provided values.
		 *Any Date values should be passed as ticks to the Jackcess.DateFromTicks helper method
		 *and the returned Date object used here.
		 */
		public void UpdateCurrentRow(Object[] values) throws Exception
		{
			cursor.updateCurrentRow(values);
			rowmap = cursor.getCurrentRow();
		}
		
		/**
		 *After calling this method, GetNextRow will return the first row in the table, 
		 */
		public void Reset()
		{
			cursor.reset();
		}
		
	}

}
